extern crate midir;
extern crate pancurses;

use std::error::Error;
use std::thread;
use std::time::{Duration, Instant};
use std::sync::{Arc, Mutex};
use std::sync::atomic::{AtomicUsize, Ordering};

use midir::os::unix::{VirtualInput, VirtualOutput};

use pancurses::{initscr, endwin, noecho, Input};

#[derive(Clone, Copy, Debug)]
enum LooperState {
    Idle,
    Recording,
    Playing
}

struct LooperInputCallback {
    output: Arc<Mutex<midir::MidiOutputConnection>>,
    state: Arc<AtomicUsize>,
    outer_start: Arc<Mutex<Instant>>,
    ts_offset: u64,
    msg_buf: Arc<Mutex<Vec<(u64, Vec<u8>)>>>,
}

struct Looper {
    virtual_input: Arc<midir::MidiInputConnection<LooperInputCallback>>,
    virtual_output: Arc<Mutex<midir::MidiOutputConnection>>,

    cur_state: Arc<AtomicUsize>,
    // (time in milliseconds, msg)
    msg_buf: Arc<Mutex<Vec<(u64, Vec<u8>)>>>,
    // for either playback or recording
    start_time: Arc<Mutex<Instant>>,
    playback_thread: Option<std::thread::Thread>,
}

impl Looper {
    const IDLE: usize = 0;
    const RECORDING: usize = 1;
    const PLAYBACK: usize = 2;

    fn new() -> Result<Looper, Box<dyn Error>> {
        let input = midir::MidiInput::new("midi looper")?;
        let output = midir::MidiOutput::new("midi looper")?;

        let cur_state = Arc::new(AtomicUsize::new(Looper::IDLE));
        let msg_buf = Arc::new(Mutex::new(Vec::new()));
        let start_time = Arc::new(Mutex::new(Instant::now()));

        let output_unboxed = output.create_virtual("looper output")?;
        let virtual_output = Arc::new(Mutex::new(output_unboxed));

        let virtual_input = input.create_virtual(
            "looper input",
            |ts, data, me| {
                {
                    let mut out = me.output.lock().unwrap();
                    out.send(data);
                }
                if me.state.load(Ordering::Acquire) == Looper::RECORDING {
                    let msg_time = Instant::now();
                    let ts_millis = ts/1000;

                    let mut buf = me.msg_buf.lock().unwrap();

                    // new recording session; find the new offset
                    if buf.len() == 0 {
                        let record_start = me.outer_start.lock().unwrap();
                        let time_since_start = msg_time.duration_since(*record_start);
                        // let's assume the user pressed their first note less than
                        // 2^64 milliseconds after starting recording
                        let time_millis = time_since_start.as_millis() as u64;
                        me.ts_offset = ts_millis - time_millis;
                    }
                    buf.push((ts_millis - me.ts_offset, data.to_vec()));
                }
            }, LooperInputCallback {
                output: virtual_output.clone(),
                state: cur_state.clone(),
                outer_start: start_time.clone(),
                ts_offset: 0,
                msg_buf: msg_buf.clone(),
            })?;

        Ok(Looper {
            virtual_input: Arc::new(virtual_input),
            virtual_output,

            cur_state: cur_state.clone(),
            msg_buf: msg_buf.clone(),
            start_time: start_time.clone(),
            playback_thread: None,
        })
    }

    fn toggle_state(&mut self) -> Result<(), Box<dyn Error>>{
        let now = Instant::now();
        let next_state = (self.cur_state.load(Ordering::Acquire) + 1) % 3;
        match next_state {
            Looper::RECORDING => {
                let mut buf = self.msg_buf.lock().unwrap();
                let mut start_time = self.start_time.lock().unwrap();
                buf.clear();
                *start_time = now;
            },
            Looper::PLAYBACK => {
                // TODO
            },
            _ => ()
        }
        self.cur_state.store(next_state, Ordering::Release);
        Ok(())
    }

    fn state(&self) -> LooperState {
        let cur_state = self.cur_state.load(Ordering::Acquire);
        match cur_state {
            0 => LooperState::Idle,
            1 => LooperState::Recording,
            2 => LooperState::Playing,
            _ => unreachable!("invalid looper state reached {}", cur_state)
        }
    }
}

fn main() {
    let mut looper = Looper::new().unwrap();

    let window = initscr();
    window.printw("============================\n");
    window.printw("======== MIDI LOOPER =======\n");
    window.printw("============================\n");
    window.printw("\n");
    window.printw("IDLE\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("start/stop = space\n");
    window.printw("save = s\n");
    window.printw("exit = Esc/Ctrl-C\n");
    window.refresh();
    window.keypad(true);
    noecho();
    loop {
        match window.getch() {
            Some(Input::Character('\x1b')) => break,
            Some(Input::Character(c)) => {
                match c {
                    ' ' => {
                        looper.toggle_state().unwrap();
                        window.mvprintw(4, 0, match looper.state() {
                            LooperState::Idle =>        "IDLE     ",
                            LooperState::Recording =>   "RECORDING",
                            LooperState::Playing =>     "PLAYING  "
                        });
                        window.mv(12, 0);
                        window.refresh();
                    },
                    's' => {
                    },
                    _ => ()
                }
            },
            Some(input) => {
                window.addstr(&format!("{:?}", input));
                window.refresh();
            },
            None => ()
        }
    }
    endwin();
}
